#include "My_mvt.h"
#include "VT.h"
#include "Burden.h"
#include "Constant.h"
#include "math.h"
#include "MathStats.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include <algorithm>
#include <time.h>
#include "Random.h"
#define MATHLIB_STANDALONE 
#include <Rmath.h>

bool VT::asymptotic = false;
bool VT::permuteBool = false;
bool VT::vcfAnnotated = true;
bool VT::preProcess = false;
String VT::chopTo = "chop_to";
int VT::chop = 20;
int VT::permuteMin = 1000;
int VT::permuteMax = 3000000;
bool VT::VTwtMB = false;
bool VT::VTwtCMC_binary = false;
bool VT::VTwtCMC_counts = true;
bool VT::noStep = false;


VT::VT(){}

VT::~VT()
{}

bool VT::GetTmpGenoMatrixPED(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;

   mac =0.0;
   numVarPerGene=0;

   tmpGenoMatrix.Dimension(0,0);
   varsIncluded.Clear();
   mafIncluded.Clear();
   Vector mafSqrIncluded;

   for(int m=0;m<group.chrList[groupNum].Length();m++)
   {
      String markerName = group.chrList[groupNum][m];
      markerName += ":";
      markerName += group.posList[groupNum][m];
      int markerid = ped.markerLookup.Integer(markerName);
      if(markerid==-1)
      {
	 continue;
      }
      int alleleCount = CountAlleles(ped, markerid);
      if(alleleCount!=2)
	 continue;

      String refAllele = ped.GetMarkerInfo(markerid)->alleleLabels[1];
      String altAllele = ped.GetMarkerInfo(markerid)->alleleLabels[2];

      String allele = refAllele + "/" + altAllele;
      if(allele != group.alleleList[groupNum][m])
      {
	 String flipAllele = altAllele + "/" + refAllele;
	 if(flipAllele != group.alleleList[groupNum][m])
	 {
	    continue;
	 }
      }
      //if marker m from group list is found then get the genotype vector
      bool XStatus = false;
      if(AutoFit::fitX)
      {
	 int found = markerName.FastFind(SVT::xLabel,0);
	 if(found!=-1)
	 {
	    if(group.posList[groupNum][m] > SVT::Xstart && group.posList[groupNum][m] < SVT::Xend)
	    {
	       XStatus=true;
	    }
	 }
      }

      int n=0,nf=0;
      double fmaf = 0.0, maf=0.0;
      Vector genotype;
      genotype.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 genotype[i]=-1.0;

      //Loop through samples according to trans.pPheno 
      int idx=0;
      for(int f=0;f<ped.familyCount;f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++)
	 {
	    if(!ped[trans.pPheno[f][p]].isGenotyped(markerid))
	    {
	       idx++;
	       continue;
	    }
	    if(ped[trans.pPheno[f][p]].markers[markerid].isHeterozygous())
	    {
	       if(XStatus)
	       {
		  if(ped[trans.pPheno[f][p]].sex==SVT::femaleLabel)
		  {
		     genotype[idx] = 1.0;
		  }
		  if(ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
		  {
		     genotype[idx] = 2.0;
		  }
	       }
	       else
	       {
		  genotype[idx] = 1.0;
	       }
	    }
	    else if(ped[trans.pPheno[f][p]].markers[markerid].isHomozygous())
	    {
	       if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(2))
	       {
		  genotype[idx] = 2.0;
	       }
	       else if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(1))
	       {
		  genotype[idx] = 0.0;
	       }
	    }
	    maf+=genotype[idx];
	    n++;
	    int founder = trans.foundersHash.Integer(ped[trans.pPheno[f][p]].pid);
	    if(founder != -1)
	    {
	       fmaf += genotype[idx];
	       nf++;
	    }
	    idx++;
	 }
      }
      maf /= 2.0*n;
      fmaf /= 2.0*nf;

      if(fmaf==0.0 && maf>0)
	 fmaf = 1.0/(2.0*nf);

      //check if maf falls into the range
      if(!group.freqOut)
      {
	 if(SVT::FounderFreq)
	 {
	    double fmaf_tmp;
	    fmaf_tmp = fmaf;
	    if(fmaf>0.5)
	       fmaf_tmp = 1.0-fmaf;
	    if(fmaf_tmp>Burden::mafMax || fmaf_tmp<Burden::mafMin)
	       continue;
	 }
	 else
	 {
	    double maf_tmp;
	    maf_tmp=maf;
	    if(maf>0.5)
	       maf_tmp = 1.0-maf;
	    if(maf_tmp>Burden::mafMax || maf_tmp<Burden::mafMin)
	       continue;
	 }
      }
      if(maf>0. && maf!=1.0)
      {
	 status=true;
      }
      else
      {
	 continue;
      }
      numVarPerGene++;

      bool flip=false;
      if(SVT::FounderFreq && fmaf>0.5)
      {
	 flip=true;
      }
      else if(maf>0.5)
      {
	 flip=true;
      }
      if(flip)
      {
	 fmaf = 1.0-fmaf;
	 maf = 1.0-maf;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       genotype[i] = 2.0-genotype[i];
	    }
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {
	    mac+=genotype[i];
	 }
      }

      if(Burden::fullResult)
      {
	 String tmp = "";
	 tmp += group.chrList[groupNum][m];
	 tmp+=":";
	 tmp+=group.posList[groupNum][m];
	 tmp+= "_";
	 if(flip)
	 {
	    tmp+=altAllele;
	    tmp+="/";
	    tmp += refAllele;
	 }
	 else
	 {
	    tmp+=refAllele;
	    tmp+="/";
	    tmp += altAllele;
	 }
	 varsIncluded.Push(tmp);
      }

      if(SVT::FounderFreq)
      {
	 mafIncluded.Push(fmaf);
	 mafSqrIncluded.Push(fmaf*fmaf);
      }
      else
      {
	 mafIncluded.Push(maf);
	 mafSqrIncluded.Push(maf*maf);
      }

      tmpGenoMatrix.GrowTo(trans.persons,numVarPerGene);

      //update tmpGenoMatrix
      for(int row=0;row<trans.persons;row++)
      {
	 tmpGenoMatrix[row][numVarPerGene-1] = genotype[row];
      }
   }
   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   if(mac<Burden::MAC)
   {
      status = false;
      return status;
   }

   return status;
}

bool VT::GetTmpGenoMatrixDosage(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;

   mac =0.0;
   numVarPerGene=0;

   tmpGenoMatrix.Dimension(0,0);
   varsIncluded.Clear();
   mafIncluded.Clear();
   Vector mafSqrIncluded;

   reader.set1BasedReadSection(group.chrList[groupNum][0].c_str(),group.posList[groupNum][0],group.posList[groupNum][group.posList[groupNum].Length()-1]+1);
   int m=0;
   while(reader.readRecord(record))
   {
      int position = record.get1BasedPosition();
      String altAllele = record.getAltStr();
      String refAllele = record.getRefStr();
      if(group.posList[groupNum][m]!=position)
      {
	 continue;
      }
      String allele = refAllele + "/" + altAllele;
      if(allele != group.alleleList[groupNum][m])
      {
	 String flipAllele = altAllele + "/" + refAllele;
	 if(flipAllele != group.alleleList[groupNum][m])
	    continue;
	 else
	    printf("ref and alt alleles are flipped!");
      }
      //If the variant match the record in groupfile then add up m to next variant. 
      //This requires varaints in group file has to be sorted.
      m++;
      if(m==group.chrList[groupNum].Length()+1)
	 break;

      int n=0,nf=0;
      double fmaf = 0.0, maf=0.0;
      VcfRecordGenotype & genoInfo = record.getGenotypeInfo();

      bool XStatus = false;
      if(AutoFit::fitX)
      {
	 String chr = group.chrList[groupNum][0].c_str();
	 if(chr==SVT::xLabel && position > SVT::Xstart && position < SVT::Xend)
	 {
	    XStatus=true;
	 }
      }

      Vector genotype;
      genotype.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 genotype[i]=-1.0;

      //Loop through samples according to trans.pPheno 
      //read in dosage into genotype vector
      int idx=0;
      for (int f = 0; f < ped.familyCount; f++)
      {
	 for(int j=0;j<trans.pPheno[f].Length();j++)
	 {
	    String sample = ped[trans.pPheno[f][j]].pid;

	    //get the sample number by sample ID from hash
	    int s = trans.sampleVCFIDHash.Integer(sample);
	    if(s!=-1)
	    {
	       const std::string * geno = genoInfo.getString("DS",s);
	       //if dosage is not missing
	       if(strcmp((*geno).c_str(),".")>0)
	       {
		  float dose = atof((*geno).c_str());
		  if(dose >=0)
		  {
		     if(XStatus && dose >0 && ped[trans.pPheno[f][j]].sex==SVT::maleLabel)
			genotype[idx] = 2;
		     else
			genotype[idx] = dose;

		     maf += genotype[idx];
		     n++;

		     //judge if this is a founder
		     int founder = trans.foundersHash.Integer(sample);
		     if(founder != -1)
		     {
			fmaf += genotype[idx];
			nf++;
		     }
		  }
	       }
	    }
	    else
	    {
	       error("FATAL ERROR! Sample %s is not included in VCF file.\n",sample.c_str());
	    }
	    idx++;
	 }
      }
      maf /= 2.0*n;
      fmaf /= 2.0*nf;

      if(fmaf==0.0 && maf>0)
	 fmaf = 1.0/(2.0*nf);

      //check if maf falls into the range
      if(!group.freqOut)
      {
	 if(SVT::FounderFreq)
	 {
	    double fmaf_tmp;
	    fmaf_tmp = fmaf;
	    if(fmaf>0.5)
	       fmaf_tmp = 1.0-fmaf;
	    if(fmaf_tmp> Burden::mafMax || fmaf_tmp<Burden::mafMin)
	       continue;
	 }
	 else
	 {
	    double maf_tmp;
	    maf_tmp=maf;
	    if(maf>0.5)
	       maf_tmp = 1.0-maf;
	    if(maf_tmp>Burden::mafMax || maf_tmp<Burden::mafMin)
	       continue;
	 }
      }
      if(maf>0. && maf!=1.0)
	 status=true;
      else
      {
	 continue;
      }
      numVarPerGene++;

      bool flip=false;
      if(SVT::FounderFreq && fmaf>0.5)
      {
	 flip=true;
      }
      else if(maf>0.5)
      {
	 flip=true;
      }
      if(flip)
      {
	 fmaf = 1.0-fmaf;
	 maf = 1.0-maf;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       genotype[i] = 2.0-genotype[i];
	    }
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {  
	    mac+=genotype[i];
	 }  
      }

      if(Burden::fullResult)
      {
	 String tmp = "";
	 tmp += group.chrList[groupNum][m-1];
	 tmp+=":";
	 tmp+=group.posList[groupNum][m-1];
	 tmp+= "_";
	 if(flip)
	 {
	    tmp+=altAllele;
	    tmp+="/";
	    tmp += refAllele;
	 }
	 else
	 {
	    tmp+=refAllele;
	    tmp+="/";
	    tmp += altAllele;
	 }
	 varsIncluded.Push(tmp);
      }

      if(SVT::FounderFreq)
      {
	 mafIncluded.Push(fmaf);
	 mafSqrIncluded.Push(fmaf*fmaf);
      }
      else
      {
	 mafIncluded.Push(maf);
	 mafSqrIncluded.Push(maf*maf);
      }

      tmpGenoMatrix.GrowTo(trans.persons,numVarPerGene);

      //update tmpGenoMatrix
      for(int row=0;row<trans.persons;row++)
      {
	 tmpGenoMatrix[row][numVarPerGene-1] = genotype[row];
      }
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   if(mac<Burden::MAC)
   {
      status = false;
      return status;
   }

   return status;
}

bool VT::GetTmpGenoMatrixVCF(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;

   varsIncluded.Clear();
   //mac is the total number of alternative alleles included
   mac = 0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;

   mafIncluded.Clear();
   Vector mafSqrIncluded;

   reader.set1BasedReadSection(group.chrList[groupNum][0].c_str(),group.posList[groupNum][0],group.posList[groupNum][group.posList[groupNum].Length()-1]+1);
   int m=0;
   while(reader.readRecord(record))
   {
      int position = record.get1BasedPosition();
      String altAllele = record.getAltStr();
      String refAllele = record.getRefStr();

      bool XStatus = false;

      if(AutoFit::fitX)
      {
	 String chr = group.chrList[groupNum][0].c_str();
	 if(chr==SVT::xLabel && position > SVT::Xstart && position < SVT::Xend)
	 {
	    XStatus=true;
	 }
      }
      if(group.posList[groupNum][m]!=position)
      {
	 continue;
      }
      else
      {
	 String allele = refAllele + "/" + altAllele;
	 if(allele != group.alleleList[groupNum][m])
	 {
	    String flipAllele = altAllele + "/" + refAllele;
	    if(flipAllele != group.alleleList[groupNum][m])
	       continue;
	    else
	    {
	       printf("ref and alt alleles are flipped!");
	    }
	 }
      }
      //If the variant match the record in groupfile then add up m to next variant. 
      //This requires varaints in group file has to be sorted.
      m++;
      if(m==group.chrList[groupNum].Length()+1)
	 break;

      double fmaf=0.0,maf=0.0;
      int n=0,nf=0;

      //Fill in genotype vector
      Vector genotype;
      genotype.Dimension(trans.persons);
      for(int i=0;i<trans.persons;i++)
	 genotype[i] = -1.0;

      int idx=0;
      for (int f = 0; f < ped.familyCount; f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++)
	 {
	    String sample = ped[trans.pPheno[f][p]].pid.c_str();

	    //get the sample number by sample ID from hash
	    int s = trans.sampleVCFIDHash.Integer(sample);
	    if(s!=-1)
	    {
	       int numGTs = record.getNumGTs(s);
	       Vector allele;

	       for(int j = 0; j < numGTs; j++)
	       {
		  allele.Push(record.getGT(s,j));
	       }
	       int bad =0;
	       for(int m=0;m<allele.Length();m++)
	       {
		  bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
	       }
	       //if no missing allele exists
	       //if no missing allele exists
	       if(bad==0)
	       {
		  int sum=0;
		  for(int i=0;i<allele.Length();i++)
		     sum+=allele[i];

		  if(XStatus && sum >0 && ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
		     genotype[idx] = 2;
		  else
		     genotype[idx] = sum;

		  maf += genotype[idx];
		  n++;
		  //judge if this is a founder
		  int founder = trans.foundersHash.Integer(sample);
		  if(founder!=-1)
		  {
		     nf++;
		     fmaf+=genotype[idx];
		  }
	       }
	    }
	    idx++;
	 }
      }
      //Adjust genotype here
      maf /= 2.0*n;
      fmaf /=2.0*nf;

      if(fmaf==0.0 && maf>0)
	 fmaf = 1.0/(2.0*nf);

      //check if maf falls into the range
      if(!group.freqOut)
      {
	 if(SVT::FounderFreq)
	 {
	    double fmaf_tmp;
	    fmaf_tmp = fmaf;
	    if(fmaf>0.5)
	       fmaf_tmp = 1.0-fmaf;
	    if(fmaf_tmp>Burden::mafMax || fmaf_tmp<Burden::mafMin)
	       continue;
	 }
	 else
	 {
	    double maf_tmp;
	    maf_tmp=maf;
	    if(maf>0.5)
	       maf_tmp = 1.0-maf;
	    if(maf_tmp>Burden::mafMax || maf_tmp<Burden::mafMin)
	       continue;
	 }
      }
      if(maf>0. && maf!=1.0)
	 status=true;
      else
      {
	 continue;
      }
      numVarPerGene++;

      bool flip=false;
      if(SVT::FounderFreq && fmaf>0.5)
      {
	 flip=true;
      }
      else if(maf>0.5)
      {
	 flip=true;
      }
      if(flip)
      {
	 fmaf = 1.0-fmaf;
	 maf = 1.0-maf;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       genotype[i] = 2.0-genotype[i];
	    }
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {  
	    mac+=genotype[i];
	 }  
      }

      if(Burden::fullResult)
      {
	 String tmp = "";
	 tmp += group.chrList[groupNum][m-1];
	 tmp+=":";
	 tmp+=group.posList[groupNum][m-1];
	 tmp+= "_";
	 if(flip)
	 {
	    tmp+=altAllele;
	    tmp+="/";
	    tmp += refAllele;
	 }
	 else
	 {
	    tmp+=refAllele;
	    tmp+="/";
	    tmp += altAllele;
	 }
	 varsIncluded.Push(tmp);
      }

      if(SVT::FounderFreq)
      {
	 mafIncluded.Push(fmaf);
	 mafSqrIncluded.Push(fmaf*fmaf);
      }
      else
      {
	 mafIncluded.Push(maf);
	 mafSqrIncluded.Push(maf*maf);
      }
      tmpGenoMatrix.GrowTo(trans.persons,numVarPerGene);

      //update tmpGenoMatrix
      for(int row=0;row<trans.persons;row++)
      {
	 tmpGenoMatrix[row][numVarPerGene-1] = genotype[row];
      }
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   if(mac<Burden::MAC)
   {
      status = false;
      return status;
   }

   return status;
}

bool VT::GetGenotypeMatrix(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;
   tmpGenoMatrix.Dimension(0,0);

   if(SVT::genoFromPed)
      status = GetTmpGenoMatrixPED(trans,ped,group,groupNum);
   if(SVT::genoFromVCF)
      status = GetTmpGenoMatrixVCF(trans,ped,group,groupNum);
   if(SVT::dosage)
      status = GetTmpGenoMatrixDosage(trans,ped,group,groupNum);

   //adjust the genotypeMatrix by its mean
   for(int col=0;col<numVarPerGene;col++)
   {
      double mean =0.0;
      int n=0;
      for(int i=0;i<trans.persons;i++)
      {
	 if(tmpGenoMatrix[i][col]!=-1.0)
	 {
	    mean+=tmpGenoMatrix[i][col];
	    n++;
	 }
      }
      mean /=n;
      for(int i=0;i<trans.persons;i++)
      {
	 if(tmpGenoMatrix[i][col]==-1.0)
	 {
	    tmpGenoMatrix[i][col] = 0.0;
	 }
	 else
	    tmpGenoMatrix[i][col] -=mean;
      }
   }

   return status;
}

bool VT::FinalizeGenoMatrix(FastTransform & trans,bool reduceDimension)
{
   //printf("Finalizing genotype matrix.\n");
   //work on mafIncluded to get unique cutoffs
   bool status = true;
   if(mafIncluded.Length()==1)
   {
      unique_maf.Copy(mafIncluded);
      transGeno.Product(trans.UD,tmpGenoMatrix);
      selectionIndicator.Dimension(unique_maf.Length(),mafIncluded.Length(),1);
      return status;
   }
   Vector cp_mafIncluded;
   cp_mafIncluded.Copy(mafIncluded);
   cp_mafIncluded.Sort();
   unique_maf.Clear();
   unique_maf.Push(cp_mafIncluded[0]);
   for(int i=1;i<cp_mafIncluded.Length();i++)
   {
      if(cp_mafIncluded[i]>unique_maf[unique_maf.Length()-1])
	 unique_maf.Push(cp_mafIncluded[i]);
   } //now unique maf cutoffs are saved in unique_maf.

   if(reduceDimension)
   {
      while(unique_maf.Length()>20)
      {
	 unique_maf.Delete(0);
      }
   }
   Matrix genotypeMatrix;
   genotypeMatrix.Dimension(trans.persons,unique_maf.Length());
   genotypeMatrix.Zero();
   selectionIndicator.Dimension(unique_maf.Length(),mafIncluded.Length(),0);

   for(int j=0;j<unique_maf.Length();j++)
   {
      for(int i=0;i<mafIncluded.Length();i++)
      {
	 if(mafIncluded[i]<=unique_maf[j])
	 {
	    selectionIndicator[j][i]=1;
	    for(int p=0;p<trans.persons;p++)
	    {
	       genotypeMatrix[p][j] += tmpGenoMatrix[p][i];
	    }
	 }
      }
   }

   /*
      if(VTwtCMC_binary)
      {
      for(int p=0;p<trans.persons;p++)
      {
      if(genotypeMatrix[p][j]>0.0)
      genotypeMatrix[p][j] = 1.0;
      }
      }
    */
   if(genotypeMatrix.cols==0)
   {
      status = false;
      return status;
   }

   transGeno.Product(trans.UD,genotypeMatrix);
   return status;
}

void VT::CalculateAsymptotic(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group)
{
   //double sigma_g2 = engine.sigma_g2Hat;
   //double sigma_e2 = engine.sigma_e2Hat;
   //double h2 = sigma_g2/(sigma_g2+sigma_e2);
   // printf("g,e,var:%g,%g,%g\n",sigma_g2,sigma_e2,trans.traitVar);
   double h2 = engine.sigma_g2Hat/trans.traitVar;

   FILE * VTout;

   String VTfilename;
   VTfilename = Burden::outputFile + traitName + ".VT.Association.txt";
   VTout = fopen(VTfilename,"wt");

   if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==1)
   {
      if(VTwtMB)
      {
	 fprintf(VTout,"\nWeight used is Madsen-Browning.\n");
      } 
      else if(VTwtCMC_binary)
      {
	 fprintf(VTout,"\nWeight used is VTwtCMC_binary.\n");
      } 
      else
      {
	 fprintf(VTout,"\nWeight used is VTwtCMC_counts.\n");
      } 
   }
   else
   {
      fprintf(VTout,"\nWeights used are: %s%s%s.\n",VTwtMB?"Madsen-Browning ":"",VTwtCMC_binary?"VTwtCMC_binary ":"",VTwtCMC_counts?"VTwtCMC_counts ":"");
   }

   fprintf(VTout,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",traitName.c_str(),trans.traitMean,trans.traitVar,h2*100.0);
   fprintf(VTout,"===================================================================\n");
   fprintf(VTout,"Gene\tNumVarIncluded\tMafCutoff\tActualThreshold\tVarsIncluded\tMAFs\tSingleVarPvalues\tSingleVarEffs\teffSize\tpvalue\n");

   int bad =0;
   Matrix rh;
   Matrix UDX_t,tmp;
   UDX_t.Transpose(trans.UDX);
   tmp.Product(trans.inv,UDX_t);
   rh.Transpose(tmp);

   for(int i=0;i<group.annoGroups.Length();i++)
   {
      if(i>1000 && i%1000==1)
	 printf("Finished analyzing %d genes.\n",((int)i/1000)*1000);
      String geneName = group.annoGroups[i];
      //printf("Testing gene %s.\n",geneName.c_str());

      if(GetGenotypeMatrix(trans,ped,group,i))
      {
	 if(Burden::fullResult)
	    CalculateSingleVarPvalue(trans,geneName,engine,tmpGenoMatrix);

	 if(!FinalizeGenoMatrix(trans,false))
	 {
	    //print output here
	    bad++;
	    if(Burden::fullResult)
	       fprintf(VTout,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    else
	       fprintf(VTout,"%s\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	    continue;
	 }
      }
      else 
      {
	 //print output here
	 bad++;
	 if(Burden::fullResult)
	    fprintf(VTout,"%s\t-\t-\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	 else
	    fprintf(VTout,"%s\t-\t-\t-\t-\t-\t-\t-\n",geneName.c_str());
	 continue;
      }

      Vector score,effSize;
      score.Dimension(transGeno.cols);
      effSize.Dimension(transGeno.cols);
      double pvalue,numerator,denominator;
      Vector denominator_2;
      denominator_2.Dimension(trans.fixedEffects);

      Matrix U;
      U.Dimension(transGeno.cols,trans.persons);

      for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
      {
	 numerator = 0.0;
	 denominator = 0.0;
	 denominator_2.Zero();
	 for(int i=0;i<trans.persons;i++)
	 {
	    numerator += residuals_original[i]*transGeno[i][cutoff];
	    denominator += transGeno[i][cutoff]*transGeno[i][cutoff];
	    for(int j=0;j<trans.fixedEffects;j++)
	    {
	       denominator_2[j] += transGeno[i][cutoff]*trans.UDX[i][j];
	    } 
	 }
	 Vector tmp; //tmp is T(sum_SiZi)*inv
	 tmp.Dimension(trans.fixedEffects);

	 for(int i=0;i<trans.fixedEffects;i++)
	 {
	    tmp[i] = trans.inv[i].InnerProduct(denominator_2);
	 }

	 denominator -= tmp.InnerProduct(denominator_2); 
	 //fill in U_ki for covariance matrix
	 for(int p=0;p<trans.persons;p++)
	 {
	    U[cutoff][p] = residuals_original[p] *(transGeno[p][cutoff] - tmp.InnerProduct(trans.UDX[p]));
	 }

	 if(denominator==0) score[cutoff]=0.0;
	 else score[cutoff] = fabs(numerator/sqrt(denominator))/sqrt(sigma2_hat);
	 effSize[cutoff] = numerator/denominator/sigma2_hat;
      }

      int cutoff = transGeno.cols;
      //get the chosen cutoff and output to .log file
      double t_max = score[0];
      int chosen_cutoff=0;
      for(int i=1;i<cutoff;i++)
      {
	 if(t_max < score[i])
	 {
	    t_max = score[i];
	    chosen_cutoff = i;
	 }
      }
      //double t_max = score.Max()/sqrt(sigma2_hat);

      String method;
      if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==1)
      {
	 if(VTwtMB) 
	    method = "MB";
	 if(VTwtCMC_binary)
	    method = "VTwtCMC_binary";
	 if(VTwtCMC_counts)
	    method = "VTwtCMC_counts";
      }
      else if(VTwtMB && VTwtCMC_binary && !VTwtCMC_counts)
      {
	 if(chosen_cutoff < length)
	    method = "MB";
	 else 
	    method = "VTwtCMC_binary";
      }
      else if(VTwtMB && VTwtCMC_counts && !VTwtCMC_binary)
      {
	 if(chosen_cutoff < length)
	    method = "MB";
	 else
	    method = "VTwtCMC_counts";
      }
      else if(VTwtCMC_counts && VTwtCMC_binary && !VTwtMB)
      {
	 if(chosen_cutoff < length)
	    method = "VTwtCMC_binary";
	 else
	    method = "VTwtCMC_counts";
      }
      else if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==3)
      {
	 if(chosen_cutoff < length)
	    method = "MB";
	 else if(chosen_cutoff < 2*length)
	    method = "VTwtCMC_binary";
	 else 
	    method = "VTwtCMC_counts";
      }

      //	    printf("chosen_cutoff is:%d\n",chosen_cutoff);
      double selected_maf = unique_maf[chosen_cutoff];

      double * lower = new double [cutoff];
      double * upper = new double [cutoff];
      double * mean = new double [cutoff];

      for(int i=0;i<cutoff;i++)
      {
	 mean[i] = 0.0;
	 lower[i] = -t_max;
	 upper[i] = t_max;
      }

      Matrix cov_U; //this is the cov matrix for U
      cov_U.Dimension(cutoff,cutoff);
      for(int r=0;r<cutoff;r++)
      {
	 for(int c=r;c<cutoff;c++)
	 {
	    cov_U[r][c] = cov_U[c][r] = U[r].InnerProduct(U[c]);
	 }
      }

      Matrix cov; //this is the cov matrix for T
      cov.Dimension(cov_U.rows,cov_U.cols);
      cov2cor(cov_U,cov);

      //Use pmvnorm to calculate the asymptotic p-value
      Vector result;
      result.Copy(pmvnorm(lower,upper,mean,cov,false));
      //printf("pmvnorm results:%g,%g,%g\n",result[0],result[1],result[2]);

      String vars="";
      for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
      {
	 if(selectionIndicator[chosen_cutoff][id]==1)
	 {
	    vars += varsIncluded[id];
	    vars += ";";
	 }
      }

      pvalue = 1.0 - result[0];
      if(pvalue==2.0)
      {
	 fprintf(VTout,"%s\t%d\t%g\t%g\t%s\t",geneName.c_str(),(int)selectionIndicator[chosen_cutoff].Sum(),Burden::mafMax,selected_maf,vars.c_str());
	 for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
	 {
	    if(selectionIndicator[chosen_cutoff][id]==1)
	       fprintf(VTout,"%g;",mafIncluded[id]);
	 }
	 fprintf(VTout,"\t");
	 for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
	 {
	    if(selectionIndicator[chosen_cutoff][id]==1)
	       fprintf(VTout,"%g;",singlevarPvalue[id]);
	 }
	 fprintf(VTout,"\t");
	 for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
	 {
	    if(selectionIndicator[chosen_cutoff][id]==1)
	       fprintf(VTout,"%g;",singlevarEff[id]);
	 }
	 fprintf(VTout,"\t%g",effSize[chosen_cutoff]);
	 fprintf(VTout,"\tERROR:CORR_MATRIX_NOT_POS_SEMI_DEF\n");
	 continue;
      }

      if(pvalue==0.0)
      {
//	 printf("using Shuang's algorithm for mvnorm prob ... ");
	 //printf("result0 is:%g\n",result[0]);
	 //TODO:reduce dimension here
	 /*
	    if(unique_maf.Length()>20)
	    FinalizeGenoMatrix(trans,true);
	  */
//	 pvalue = CalculatePvalue(score,cov,t_max);
//	 printf("completed!\n");

      }

      fprintf(VTout,"%s\t%d\t%g\t%g\t%s\t",geneName.c_str(),(int)selectionIndicator[chosen_cutoff].Sum(),Burden::mafMax,selected_maf,vars.c_str());
      for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
      {
	 if(selectionIndicator[chosen_cutoff][id]==1)
	    fprintf(VTout,"%g;",mafIncluded[id]);
      }
      fprintf(VTout,"\t");
      for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
      {
	 if(selectionIndicator[chosen_cutoff][id]==1)
	    fprintf(VTout,"%g;",singlevarPvalue[id]);
      }
      fprintf(VTout,"\t");
      for(int id=0;id<selectionIndicator[chosen_cutoff].Length();id++)
      {
	 if(selectionIndicator[chosen_cutoff][id]==1)
	    fprintf(VTout,"%g;",singlevarEff[id]);
      }
      fprintf(VTout,"\t%g",effSize[chosen_cutoff]);
      fprintf(VTout,"\t%g\n",pvalue);

      if(lower) delete [] lower;
      if(upper) delete [] upper;
      if(mean) delete [] mean;
   }

   fprintf(VTout,"===================================================================\n");
   fprintf(VTout,"\n%d %s tested.\n",group.annoGroups.Length(),group.annoGroups.Length()>1?"genes were":"gene was");
   fprintf(VTout,"There %s %d %s either had no rare allele in the qualified sample or did not pass the collapsing criterion.\n",bad>1?"are":"is", bad,bad>1?"genes":"gene");

   fclose(VTout);
}

double VT::CalculatePvalue(Vector & score,Matrix & cov,double t_max)
{
   double pvalue=0.0;
   int dim = score.Length();
   double chisq = t_max * t_max;

   StringDoubleHash jointProbHash;

   if(dim==1)
   {
      pvalue = chidist(chisq,1);
      //printf("pvalue is: %g\n",pvalue);
      return pvalue;
   }
   double uni = chidist(chisq,1);
   pvalue += dim*uni;
   //printf("pvalue so far is:%g\n",pvalue);

   //Initializations for combination
   vector<int> indx;
   vector<int> alpha (dim,0);
   for(int i=0;i<dim;i++)
      alpha[i] = i;
   int n=dim;

   for(int r=2;r<=dim;r++)
   {
      int j=r;
      int k=r;

      Vector par,comb;
      par.Dimension(2);

      for(int twk=j;twk<=k;twk++)
      {
	 int r=twk;
	 bool done=true;
	 for(int iwk=0;iwk<r;iwk++)indx.push_back(iwk);
	 while(done)
	 {
	    done=false;
	    for(int owk=0;owk<r;owk++)
	    {
	       //A new combination has been generated!
	       comb.Push(alpha[indx[owk]]);
	    }
	    /*
	       printf("New combination:\n");
	       for(int i=0;i<comb.Length();i++)
	       printf("%g ",comb[i]);
	       printf("\n");
	     */
	    //Calculate P(A|BC)
	    par.Copy(GetConditionalDist(score,cov,comb));
	    //printf("pars are:%g %g\n",par[0],par[1]);
	    double chisq,condProb,prob;
	    if(par[1]==0.0)
	    {
	       condProb = 0.0;
	    }
	    else 
	    {
	       chisq = (t_max-par[0])*(t_max-par[0])/par[1];
	       if(chisq<0.0)
		  chisq = -chisq;
	       condProb = chidist(chisq,1);
	    }
	    //printf("conditional prob is:%g\n",condProb);
	    String hashKey;

	    if(r==2)
	    {
	       int tmp = comb[0];
	       hashKey += tmp;
	       tmp = comb[1];
	       hashKey += tmp;
	       //	       printf("hashKey is: %s\n",hashKey.c_str());
	       prob = condProb*uni;	       
	       //prob = exp(log(condProb) + log(uni));
	       //printf("joint prob is:%g\n",prob);
	       jointProbHash.SetDouble(hashKey,prob);
	       hashKey.Clear();
	    }
	    else
	    {
	       for(int i=1;i<r;i++)
	       {
		  int tmp = comb[i];
		  hashKey += tmp;
	       }
	       //	       printf("hashKey is: %s\n",hashKey.c_str());
	       prob = jointProbHash.Double(hashKey);
	       //printf("joint prob is:%g\n",prob);
	       //if(prob == _NAN_)
	       //{
	       //	  printf("Error! Joint probability %s is not hashed in properly.\n",hashKey);
	       //     }
	       prob *= condProb; //This is the new joint prob.
	       String newKey;
	       int tmp = comb[0];
	       newKey += tmp;
	       newKey += hashKey;
	       //	       printf("newKey is: %s\n",newKey.c_str());
	       jointProbHash.SetDouble(newKey,prob);
	       hashKey.Clear();
	    }
	    pvalue -= prob;
	    //printf("pvalue so far is:%g\n",pvalue);
	    comb.Clear();
	    for(int iwk=r-1;iwk>=0;iwk--)
	    {
	       if(indx[iwk]<=(n-1)-(r-iwk))
	       {
		  indx[iwk]++;
		  for(int swk=iwk+1;swk<r;swk++)
		  {
		     indx[swk]=indx[swk-1]+1;
		  }
		  iwk=-1;
		  done=true;
	       }
	    }
	 }
	 indx.clear();
      }
   }
   // printf("pvalue is:%g\n",pvalue);
   return pvalue;
}

//Get P(A|BC) where A is the first element in comb. Comb contains index only.
Vector VT::GetConditionalDist(Vector & score,Matrix & cov,Vector & comb)
{
   Vector result;
   result.Dimension(2);
   result.Zero();

   //chop score vector and sigma matrix
   Vector mu2;
   Matrix sub_cov;
   int dim = comb.Length()-1;
   mu2.Dimension(dim);
   sub_cov.Dimension(dim,dim);
   for(int i=0;i<dim;i++)
   {
      int idx1 = comb[i+1];
      mu2[i] = score[idx1];
      for(int j=0;j<dim;j++)
      {
	 int idx2 = comb[j+1];
	 sub_cov[i][j] = cov[idx1][idx2];
      }
   }
   /*printf("sub_cov is:\n");
     for(int i=0;i<dim;i++)
     {
     for(int j=0;j<dim;j++)
     {
     printf("%g ",sub_cov[i][j]);
     }
     printf("\n");
     }*/

   Matrix inv;
   inv.Copy(sub_cov);
   SVD svd;
   svd.InvertInPlace(inv);
   Vector sigma12;
   sigma12.Dimension(dim);
   for(int i=0;i<dim;i++)
   {
      int idx1 = comb[0];
      int idx2 = comb[i+1];
      sigma12[i] = cov[idx1][idx2];
   }
   /*
      printf("sigma12 is:\n");
      for(int i=0;i<dim;i++)
      printf("%g ",sigma12[i]);
      printf("\n");
    */
   Vector tmp;
   tmp.Dimension(dim);
   tmp.Zero();
   for(int i=0;i<dim;i++)
   {
      tmp[i] += sigma12.InnerProduct(inv[i]);
   }

   //Calculate Mu and Sigma here
   result[0] = tmp.InnerProduct(mu2);
   result[1] = 1.0 - tmp.InnerProduct(sigma12); 

   if(result[1]<0)
      result[1] = fabs(result[1]);

   return result;
} 

void VT::Run(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group)
{
   if(!vcfAnnotated)
   {
      error("\nERROR! VCF file needs to be annotated in INFO field.\n");
   }

   if(asymptotic || permuteBool)
   {
      //Fill in UDX, UD, and UDY
      //trans.FinalizeTransform(engine.sigma_g2Hat, engine.sigma_e2Hat);

      if(!VT::VTwtMB && !VT::VTwtCMC_binary && !VT::VTwtCMC_counts)
      {
	 VTwtCMC_binary=true;
	 printf("Reminder: weight function was specified for Variable-threshold test. VTwtCMC_binary weight will be use in the following VT tests.\n");
      }

      fixedEff.Dimension(trans.persons);
      //      beta_hat.Dimension(trans.fixedEffects);
      //    engine.EstimateAnalyticalBeta(trans.UDY,trans, beta_hat);
      for(int i=0;i<trans.persons;i++)
      {
	 fixedEff[i] = trans.UDX[i].InnerProduct(engine.betaHat);
      }
      residuals_original.Copy(trans.UDY);
      residuals_original.Subtract(fixedEff);
      sigma2_hat=1.0;
      //sigma2_hat = residuals_original.SumSquares()/trans.persons;

      if(SVT::dosage || SVT::genoFromVCF)
      {
	 reader.open(GroupFromAnnotation::vcfInput,header);
	 reader.readVcfIndex();
      }
      if(asymptotic)
      {
	 printf("\nStart doing VT tests using asmptotic method ...\n");
	 CalculateAsymptotic(ped,trans,engine,traitName,group);
	 printf("Finished doing VT tests.\n");
      }
   }
}

void VT::CalculateSingleVarPvalue(FastTransform & trans,String & geneName,FastFit & engine,Matrix & genotypeSKAT)
{
   //printf("Working on singlevar on gene %s.\n",geneName.c_str());
   Matrix transGeno;
   transGeno.Product(trans.UD,genotypeSKAT);

   // Calculate singlevariant pvalues
   singlevarPvalue.Dimension(genotypeSKAT.cols,0.0);
   singlevarEff.Dimension(genotypeSKAT.cols,0.0);
   for(int m=0;m<genotypeSKAT.cols;m++)
   {
      Vector geno;
      geno.Dimension(trans.persons);
      for(int p =0;p<trans.persons;p++)
      {
	 geno[p] = transGeno[p][m];
      }
      double numerator = residuals_original.InnerProduct(geno);
      double denominator = geno.InnerProduct(geno);
      double pvalue =  pchisq(numerator*numerator/denominator,1,0,0);
      //double pvalue =  chidist(numerator*numerator/denominator,1);
      singlevarPvalue[m] = pvalue;
      singlevarEff[m] = numerator/denominator;
   }
   //printf("finished working on singlevar on gene %s.\n",geneName.c_str());
}
